# Input generation of vasp input files using config files and templates

import os
import sys
import configparser as cp


def generate_input(incar_template):
    '''
    Generates the INCAR, and POTCAR files used to perform VASP calculations.

    Parameters
    ----------
    incar_template : str 

    Configuration files
    -------------------
    The incar_template file is a configuration file that contains the parameters for the INCAR file.
    The configuration file is a standard .ini file. The parameters are the same as those in the INCAR file.
    The parameters are read in as strings, and are converted to the appropriate type when the INCAR file is generated.
    The configuration files can be generated using the generate_config program.

    NOTE
    ----
    This program requires the use of a HOST configuration file.
    The HOST configuration file contains the information about the host machine.
    This includes:
        - The hostname
        - The architecture of the host machine and the subprocess it uses
        - The location of the POTPAW pseudopotential files on the host machine
    '''

    config = cp.ConfigParser()
    config.read('host.ini')
    hostname = config['host']['hostname']
    architecture = config['host']['architecture']
    potpaw_location = config['host']['potpaw']
    config = cp.ConfigParser()
    config.read(incar_template)
    with open('INCAR', 'w') as f:
        for section in config.sections():
            for key in config[section]:
                key = key.upper()
                f.write(f'''{key} = {config[section][key]}
''')
    with open('POTCAR','a') as f:
        f.truncate(0)
        with open('POSCAR','r') as g:
            poscar = []
            for lines in g:
                stripped_lines = lines.strip()
                poscar.append(stripped_lines)
            atoms = poscar[5].split()
            potpaw = ['H', 'He', 'Li_sv', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na_pv', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K_sv', 'Ca_sv', 'Sc_sv', 'Ti_sv', 'V_sv', 'Cr_pv', 'Mn_pv', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga_d', 'Ge_d', 'As', 'Se', 'Br', 'Kr', 'Rb_sv', 'Sr_sv', 'Y_sv', 'Zr_sv', 'Nb_sv', 'Mo_sv', 'Tc_pv', 'Ru_pv', 'Rh_pv', 'Pd', 'Ag', 'Cd', 'In_d', 'Sn_d', 'Sb', 'Te', 'I', 'Xe', 'Cs_sv', 'Ba_sv', 'La', 'Ce_3', 'Nd_3', 'Pm_3', 'Sm_3', 'Eu_2', 'Gd_3', 'Tb_3', 'Dy_3', 'Ho_3', 'Er_3', 'Tm_3', 'Yb_2', 'Lu_3', 'Hf_pv', 'Ta_pv', 'W_sv', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl_d', 'Pb_d', 'Bi_d', 'Po_d', 'At', 'Rn', 'Fr_sv', 'Ra_sv', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm']
            potpaw_f = []
            for i in range(len(atoms)):
                for j in potpaw:
                    k = j.split('_')
                    if atoms[i] == k[0]:
                            potpaw_f.append(j)
            cwd = os.getcwd()
            os.system('cd ')
            os.system('module load vasp')
            os.chdir(potpaw_location)
            for i in potpaw_f:
                os.chdir(f'{i}')
                with open('POTCAR','r') as d:
                    for line in d:
                        f.write(line)
                os.chdir('../')
            os.chdir(cwd)

def generate_job(job_template, title):
    '''
    Generates the job submission script for the VASP calculation.

    Parameters
    ----------
    job_template : str
        The name of the job template file.
    title : str 

    Configuration files
    -------------------
    The job_template file is a configuration file that contains the parameters for the job submission script.
    The configuration file is a standard .ini file. The parameters are the same as those in the job submission script.
    The parameters are used to generate the job submission script based upon the subprocess 

    NOTE
    ----
    This program requires the use of a HOST configuration file.
    The HOST configuration file contains the information about the host machine.
    This includes:
        - The hostname
        - The architecture of the host machine and the subprocess it uses
        - The location of the POTPAW pseudopotential files on the host machine
    '''

    config = cp.ConfigParser()
    config.read('host.ini')
    hostname = config['host']['hostname']
    architecture = config['host']['architecture']
    potpaw_location = config['host']['potpaw']
    config = cp.ConfigParser()
    config.read(job_template)
    version = config['job']['version']
    nodes = config['job']['nodes']
    cores = config['job']['cores']
    module_location = config['job']['module_location']
    vasp_type = config['job']['vasp_type']
    if architecture == 'slurm':
        with open('submit.sh','w') as f:
            f.truncate(0)
            f.write(f'''#!/bin/bash
#SBATCH -p {nodes}       
#SBATCH -n {cores}
#SBATCH --job-name="{title}"

echo "Starting run at: `date`"

module load {module_location}
module load anaconda3/2020.07

# Call VASP directly.

mpirun -np {cores} {vasp_type}

echo "Job finished with exit code $? at: `date`"
''')
      
        with open('slurm_Opt.sh','r') as f:
            opt = f.read()
        with open('Opt.sh','w') as f:
            f.write(opt)

    elif architecture == 'sqe':
         if nodes == 'multicore':
            with open('submit.sh','w') as f:
                f.truncate(0)
                f.write(f'''#!/bin/bash --login
#$ -cwd             # Job will run from the current directory
#$ -pe smp.pe {cores}
#$ -N '{title}'

echo "Starting run at: `date`"

module load {module_location}

mpirun -n $NSLOTS {vasp_type}

echo "Job finished with exit code $? at: `date`"
''')
         if nodes == 'multinode':
             with open('submit.sh','w') as f:
                 f.truncate(0)
                 f.write(f'''#!/bin/bash --login
#$ -cwd                       # Job will run from the current directory
#$ -pe mpi-24-ib.pe {cores}
#$ -N '{title}'

module load {module_location}

mpirun -n $NSLOTS {vasp_type}

echo "Job finished with exit code $? at: `date`"               
''')
         with open('sge_Opt.sh','r') as f:
             opt = f.read()
         with open('Opt.sh','w') as f:
             f.write(opt)
    elif architecture == 'local':
        with open('submit.sh','w') as f:
            f.truncate(0)
            f.write(f'''nohup mpirun -n {cores} {vasp_type} > output01.txt &''')

    else:
        print('Architecture not supported')
    




# a module that will write configuration scripts for the user to run to create templates for their calculations
# The required configuration filesn are: input, job, and host

def generate_config_script(config):
    """
    Generate a configuration script for the user to run to create templates for their calculations.
    The template files (input, job, and host) are required for the user to generate input and job submission scripts.
    """
    if config == 'input':
        with open('input_config.py', 'w') as f:
            f.write('''
import configparser as cp
"""
A configuration file for the input script.
The configuation below is an example of how to format a configuation file.

IMPORTANT
--------
so not edit ['!general'], ['!calc_type'], ['!functional'] or ['!encut'] these are read in the function,
only change the parameters in the braces.
"""
config = cp.ConfigParser()
config['!general'] = {
        'PREC' : 'ACCURATE',
        'LREAL' : '.FALSE.',
        'LASPH' : '.TRUE.',
        'ISMEAR' : 0,
        'SIGMA' : 0.1,
        'NELM' : 100,
        'NELMIN' : 4,
        'NCORE' : 4,
        'EDIFF' : 1e-8,
        'EDIFFG' : -1E-2,}

config['!calc_type'] = {
        'IBRION' : 2,
        'NSW' : 100,
        'ISIF' : 4,
        'POTIM' : 0.5,
        'LWAVE' : '.FALSE.',
        'LCHARG' : '.FALSE.',
        'LORBIT' : 11,}

config['!functional'] = {
        'GGA' : 'PE',}

config['!encut'] = {
        'ENCUT' : 500,}


with open('incar.ini', 'w') as configfile:
    config.write(configfile)
''')
    elif config == 'job':
        with open('job_config.py', 'w') as f:
            f.write('''
import configparser as cp
"""
A configuration file for the input script.
The configuation below is an example of how to format a configuation file.

IMPORTANT
--------
so not edit ['job'] this is read in the function,
only change the parameters in the braces.
"""

config = cp.ConfigParser()
config['job'] = {
        'nodes' : 'mutlicore',
        'cores' : 16,
        'module_location' : 'apps/intel-17.0/vasp/6.1.2',
        'vasp_type' : 'vasp_gam'}

with open('job.ini', 'w') as configfile:
    config.write(configfile)
''')
    elif config == 'host':
        with open('host_config.py', 'w') as f:
            f.write('''
import configparser as cp
"""
A configuration file for the input script.
The configuation below is an example of how to format a configuation file.

IMPORTANT
--------
so not edit ['host'] this is read in the function,
only change the parameters in the braces.
"""

config = cp.ConfigParser()
config['host'] = {
        # The name of the host
        'hostname' : 'localhost',
        # The architecture of the host (e.g. slurm, sqe, local)
        'architecture' : 'local',
        # The loacation of the potpaw psuedopotential files 
        'potpaw' : '/home/username/potpaw',}

with open('host.ini', 'w') as configfile:
    config.write(configfile)
''')




# The following module allows for the transformation of files for use and analysis in vasp and other codes such as molcas
# The module will include the transfor from .cif file to poscar file and then poscar to .xyz file 

# Imports
import numpy as np
import os
from gemmi import cif

# Functions
def convert_cif(cif_file):
    """
    Converts a .cif file to a POSCAR file for use in vasp.
    KNOWN BUGS
    ----------
    - The program doesnt work for .cif files containig "(number)" in the positions of the atoms and the lattice vectors
    - View the .cif file in a text editor to see if this is the case

    Parameters
    ----------
    cif_file : str 

    Returns
    -------
    POSCAR 
    """

    cif_doc =  cif.read_file(cif_file)
    cif_block = cif_doc.sole_block()
    name = cif_file.strip(".cif")
    # Get the lattice vectors
    a = cif_block.find_value("_cell_length_a")
    b = cif_block.find_value("_cell_length_b")
    c = cif_block.find_value("_cell_length_c")
    a, b, c = float(a), float(b), float(c)
    alpha = cif_block.find_value("_cell_angle_alpha")
    beta = cif_block.find_value("_cell_angle_beta")
    gamma = cif_block.find_value("_cell_angle_gamma")
    alpha, beta, gamma = float(alpha), float(beta), float(gamma)
    alpha, beta, gamma = np.radians(alpha), np.radians(beta), np.radians(gamma)
    # Get the atom positions
    xpos, ypos, zpos = list(cif_block.find_loop("_atom_site_fract_x")), list(cif_block.find_loop("_atom_site_fract_y")), list(cif_block.find_loop("_atom_site_fract_z"))
    # Get the atom labels
    atom_labels = list(cif_block.find_loop("_atom_site_label"))
    # Get the atom types
    atom_types = list(cif_block.find_loop("_atom_site_type_symbol"))
    
    # Transform the lattice vectors and angles into a matix
    a_vector = np.array([a, 0, 0])
    b_vector = np.array([b*np.cos(gamma), b*np.sin(gamma), 0])
    c_vector = np.array([np.cos(beta), c*((np.cos(alpha) - np.cos(beta)*np.cos(gamma))/np.sin(gamma)), c*np.sqrt(1 - np.cos(beta)**2 - ((np.cos(alpha) - np.cos(beta)*np.cos(gamma))/np.sin(gamma))**2)])

    # Create the atom position matirx
    position_matrix = [] 
    for i in zip(xpos, ypos, zpos):
        position = np.array([float(i[0]), float(i[1]), float(i[2])])
        position_matrix.append(position)
    position_matrix = np.array(position_matrix)
    
    # Create a list of tuples containing the atom and the number of that atom (atom, number)
    atom_list = []
    for i in atom_types:
        if i not in atom_list:
            atom_list.append(i)
    atom_list = sorted(atom_list)
    atom_number_list = []
    for i in atom_list:
        atom_number_list.append(atom_types.count(i))
    atom_number_list = tuple(atom_number_list)
    atom_list = tuple(atom_list)
    atom_list = list(zip(atom_list, atom_number_list))

    atom_list_string = ""
    for i in atom_list:
        atom_list_string += i[0] + " "
    atom_number_list_string = ""
    for i in atom_list:
        atom_number_list_string += str(i[1]) + " "

    # Generate the POSCAR file
    with open('POSCAR', 'w') as f:
        f.write(f'''{name}
1.0
        {a_vector[0]} {a_vector[1]} {a_vector[2]}
        {b_vector[0]} {b_vector[1]} {b_vector[2]}
        {c_vector[0]} {c_vector[1]} {c_vector[2]}
{atom_list_string}
  {atom_number_list_string}
Direct
''')
        for i in position_matrix:
            f.write(f"    {i[0]}     {i[1]}      {i[2]}\n")


def convert_xyz():
    """
    Converts a POSCAR file to a .xyz file for use in molcas

    Parameters
    ----------
    None

    Returns
    -------
    .xyz file
    """
    with open('POSCAR','r') as f:
            poscar = []
            for lines in f:
                stripped_lines = lines.strip()
                split_lines = stripped_lines.split()
                poscar.append(split_lines)
    lattice_matrix = np.float_(poscar[2:5])
    frac_coord = np.float_(poscar[8:])
    name = ''.join(poscar[0])
    cart_coord = []
    for i in range(len(frac_coord)):
        cart_coord.append(np.matmul(frac_coord[i],lattice_matrix))
    element = poscar[5]
    element_num = poscar[6]
    element_num = [int(x) for x in element_num]
    cumsum_element = list(np.cumsum(element_num))
    element_list =[]
    for i in range(len(element)):
        element_list.append(f'{element[i]} '*int(element_num[i]))
    for i in range(len(element_list)):
        element_list[i] = ' '.join(element_list[i].split())
    element_list = ' '.join(element_list)
    element_list = element_list.split()

    with open('POSCAR.xyz','w') as f:
        f.write(f'''{len(cart_coord)}
{name}
''')
        for i in range(len(cart_coord)):
            f.write(f'''{element_list[i]}    {float(cart_coord[i][0])}    {float(cart_coord[i][1])}    {float(cart_coord[i][2])}   
''')  




# A module for plotting the density of states and projected density of states for a given material using the DOSCAR file from vasp
# Imports
import matplotlib.pyplot as plt
import numpy as np

# Functions
def plot_dos(dos_type, orbitals):
    """
    Plots the density of states for a given material
    
    Parameters
    ----------
    dos_type : str  
    orbitals : str

    Returns
    -------
    Plot of density of states
    """
    
    DOSCAR = []
    with open('DOSCAR','r') as f:
        for lines in f:
            stripped_lines = lines.strip()
            split_lines = stripped_lines.split()
            DOSCAR.append(split_lines)
    
    len_dos = []
    for i in range(len(DOSCAR)):
        if DOSCAR[5][0] in DOSCAR[i]:
            len_dos.append(i)

    
    fig,ax = plt.subplots()

    if dos_type == 'total':
        # TOTAL DOS
        dos_data = np.float_(DOSCAR[len_dos[0]+1:len_dos[1]])
        energy = dos_data[:,0]
        dos = dos_data[:,1]
        ax.plot(energy,dos,'r-')

    if dos_type == 'orbital':
        del len_dos[0]

        # PLOT d orbitals
        if orbitals == 'd': 
            dos_data = np.float_(DOSCAR[len_dos[1]+1:len_dos[2]]) #Finds location of the selected atom - needs extra code to do the location (use POSCAR for atoms)
            energy = dos_data[:,0]
            dos = dos_data[:,6]
            ax.plot(energy,dos,'r-',label='dyz')
            dos = dos_data[:,8]
            ax.plot(energy,dos,'b-',label=f'dxz')
            dos = dos_data[:,5]
            ax.plot(energy,dos,'g-',label='dxy')
            dos = dos_data[:,4]
            ax.plot(energy,dos,'c-',label='$dx^{2}-y^{2}$')
            dos = dos_data[:,7]
            ax.plot(energy,dos,'m-',label=f'$dz^2$ ')   

        #PLOT f orbitals
        if orbitals == 'f':
            dos_data = np.float_(DOSCAR[len_dos[1]+1:len_dos[2]]) #Finds location of the selected atom - needs extra code to do the location (use POSCAR for atoms)
            energy = dos_data[:,0]
            dos = dos_data[:,9]
            ax.plot(energy,dos,'r-',label='f1')
            dos = dos_data[:,10]
            ax.plot(energy,dos,'b-',label='f2')
            dos = dos_data[:11]
            ax.plot(energy,dos,'g-',label='f3')
            dos = dos_data[:,12]
            ax.plot(energy,dos,'c-',label='f4')
            dos = dos_data[:,13]
            ax.plot(energy,dos,'m-',label='f5')
            dos = dos_data[:,14]
            ax.plot(energy,dos,'m-',label='f6')
            dos = dos_data[:,15]
            ax.plot(energy,dos,'m-',label='f7')
    

    ax.set_xlabel('Energy/ eV')
    ax.set_ylabel('Density of states')
    ax.grid(color = 'grey', linestyle = '--')
    ax.legend()
    fig.tight_layout()
    #plt.show()
    plt.savefig('DOS.png')
